# Copyright (c) Qotto, 2021

"""
Django integration utilities

Utility functions to integrate the eventy protocol in django apps
"""
import logging
from typing import Callable

import django.http as vendor_django_http

import eventy.config.django
from eventy.trace_id import local_trace
from eventy.trace_id.generator import gen_trace_id

logger = logging.getLogger(__name__)


def django_trace_middleware(
    get_response: Callable[[vendor_django_http.HttpRequest], vendor_django_http.HttpResponse]
) -> Callable:
    """
    Middleware to extract and propagate correlation_id, user_id, and generate request_id.

    Log each access, except for health check on ``eventy.config.django.DJANGO_ACCESS_HEALTH_ROUTE``
    if ``eventy.config.django.DJANGO_ACCESS_DISABLE_HEALTH_LOGGING`` is set to True

    :param get_response: Actual response logic
    :return: Modified response
    """

    def _middleware(request: vendor_django_http.HttpRequest) -> vendor_django_http.HttpResponse:
        # test if access should be logged
        is_health_check: bool = request.method == 'GET' and request.path == eventy.config.django.DJANGO_ACCESS_HEALTH_ROUTE
        log_access: bool = not (is_health_check and eventy.config.django.DJANGO_ACCESS_DISABLE_HEALTH_LOGGING)

        # fetch or generate correlation_id
        correlation_id = None
        request_length = None
        user_id = None
        headers = getattr(request, 'headers', None)
        if not headers:
            # Django 1
            headers = getattr(request, 'META', {})
        for header_key, header_value in headers.items():
            # Django can change case, and Django 1 META adds HTTP_ prefix
            header_key = header_key.lower().replace('http_', '').replace('_', '-')
            if header_key == 'x-correlation-id':
                correlation_id = header_value
            if header_key == 'x-user-id':
                user_id = header_value
            if header_key == 'content-length':
                request_length = header_value
        with local_trace(
            correlation_id=correlation_id or gen_trace_id(f'{request.method}_{request.path}'),
            user_id=user_id or '',
            request_id=gen_trace_id(f'{request.method}_{request.path}'),
        ):

            # log request
            if log_access:
                logger.info(f"Request: {request.method} {request.path} ({request_length or '0'})")

            # get response
            response = get_response(request)

            # log response
            if log_access:
                response_length = 'N/A'
                try:
                    response_length = str(len(response.content))
                except Exception:
                    pass  # e.g. FileResponse
                logger.info(f'Response: {request.method} {request.path} {response.status_code} ({response_length})')

            return response

    logger.info("Eventy for Django middleware initialized.")
    return _middleware



# Copyright (c) Qotto, 2021

"""
Trace IDs local context manager
"""

from contextlib import contextmanager
from contextvars import Token
from typing import Optional

from eventy.trace_id import correlation_id_var, request_id_var, user_id_var

__all__ = [
    'local_trace',
]


@contextmanager
def local_trace(
    correlation_id: Optional[str] = None,
    request_id: Optional[str] = None,
    user_id: Optional[str] = None,
):
    """
    Local trace ids to use in with statements
    """
    # keep reset tokens
    correlation_id_token: Optional[Token]
    if correlation_id:
        correlation_id_token = correlation_id_var.set(correlation_id)
    else:
        correlation_id_token = None
    request_id_token: Optional[Token]
    if request_id:
        request_id_token = request_id_var.set(request_id)
    else:
        request_id_token = None
    user_id_token: Optional[Token]
    if user_id:
        user_id_token = user_id_var.set(user_id)
    else:
        user_id_token = None

    # yield inside the with statement
    try:
        yield correlation_id, request_id, user_id

    # reset trace ids
    finally:
        if correlation_id_token:
            correlation_id_var.reset(correlation_id_token)
        if request_id_token:
            request_id_var.reset(request_id_token)
        if user_id_token:
            user_id_var.reset(user_id_token)

    # exceptions are propagated



# Copyright (c) Qotto, 2021

def _merge_docs(eventy_object, vendor_object) -> None:
    eventy_doc: str = getattr(eventy_object, '__doc__', '')
    vendor_doc: str = getattr(vendor_object, '__doc__', '')
    if vendor_doc:
        vendor_doc = f'\n{vendor_doc}'
    merged_doc = eventy_doc + vendor_doc
    setattr(eventy_object, '__doc__', merged_doc)



# Copyright (c) Qotto, 2021

"""
Eventy celery integration configuration
"""

__all__ = [
    'CELERY_AUTO_FORGET_ON_GET',
]

CELERY_AUTO_FORGET_ON_GET: bool = False
"""
Use custom AsyncResult class to automatically forget tasks after get()
"""



# Copyright (c) Qotto, 2021

"""
Record serialization package
"""

from typing import Any

from eventy.record import Record

__all__ = [
    'RecordSerializer',
]


class RecordSerializer:
    """
    Record serializer interface
    """

    def encode(self, record: Record) -> Any:
        """
        Encode a record

        Raises:
            SerializationError
        """
        raise NotImplementedError

    def decode(self, encoded: Any) -> Record:
        """
        Decode a record

        Raises:
            SerializationError
        """
        raise NotImplementedError



# Copyright (c) Qotto, 2021

"""
Include support for context bound trace ids

* :obj:`context_vars.correlation_id_var` a Correlation ID to correlate multiple services
  involved in the same business operation
* :obj:`context_vars.request_id_var` a Request ID to identify a single request on a service

Include support to generate trace ids

* :obj:`generator.gen_trace_id` trace id generator from a function
* :obj:`local.local_trace` local context bound trace id setter
"""

from eventy.trace_id.context_vars import correlation_id_var, request_id_var, user_id_var
from eventy.trace_id.generator import gen_trace_id
from eventy.trace_id.local import local_trace

__all__ = [
    'correlation_id_var',
    'request_id_var',
    'user_id_var',
    'gen_trace_id',
    'local_trace',
]



# Copyright (c) Qotto, 2021

from __future__ import annotations

import logging
import threading
import time
from threading import Thread
from typing import Optional, Iterable, Callable

from eventy.messaging.agent import Agent, Handler, Guarantee
from eventy.messaging.service import Service
from eventy.messaging.store import Cursor, RecordStore, RecordWriteStore, RecordStoreWithInitialization
from eventy.record import Record, Response, RecordType
from eventy.trace_id.local import local_trace

__all__ = [
    'EventyApp',
]

logger = logging.getLogger(__name__)


class EventyApp:
    """
    EventyApp is the messaging application.
    """

    def __init__(
        self,
        record_store: RecordStore,
        app_service: Service,
        ext_services: Iterable[Service] = None,
        agents: Iterable[Agent] = None,
        handlers: Iterable[Handler] = None,
        read_batch_size: int = 1,
        read_timeout_ms: Optional[int] = None,
        read_wait_time_ms: Optional[int] = None,
        write_timeout_ms: Optional[int] = None,
        ack_timeout_ms: Optional[int] = None,
        commit_timeout_ms: Optional[int] = None,
    ):
        """
        Create and initialize EventyApp.

        The record store will be initialized if needed.
        """
        self._record_store = record_store
        self._app_service = app_service
        self._ext_services = ext_services or []
        self._all_handlers = list()
        if handlers is not None:
            for handler in handlers:
                self._all_handlers.append(handler)
        if agents is not None:
            for agent in agents:
                for handler in agent.handlers:
                    self._all_handlers.append(handler)

        self._read_batch_size = read_batch_size
        self._read_timeout_ms = read_timeout_ms
        self._write_timeout_ms = write_timeout_ms
        self._ack_timeout_ms = ack_timeout_ms
        self._commit_timeout_ms = commit_timeout_ms
        self._read_wait_time_ms = read_wait_time_ms or 0

        self._write_store = _AppRecordWriteStore(self)

        self._cancelled = False

        self._initialize()

    @property
    def write_store(self) -> RecordWriteStore:
        """
        The write store to write records directly without transaction management.
        """
        return self._write_store

    def _initialize(self) -> None:

        # Initialize: register topics to store if needed
        logger.info(f"Initializing EventyApp {self._app_service.name}.")
        if self._record_store.assignments:
            logger.info(f"RecordStore already has manual assignments. Will not auto register topics.")
        else:
            topics_to_register = self._get_topics_to_register()
            logger.info(f"Registering {len(topics_to_register)} topics: {', '.join(topics_to_register)}.")
            for topic in topics_to_register:
                self._record_store.register_topic(topic, Cursor.ACKNOWLEDGED)

        # Initialize: initialize store if needed
        if isinstance(self._record_store, RecordStoreWithInitialization):
            logger.info(f"Initializing RecordStore.")
            self._record_store.initialize()

        # Cancel immediately if no topics to read from
        if not self._record_store.topics and not self._record_store.assignments:
            logger.info(f"No topics to read from, cancelling app.")
            self.cancel()

    def run_until_cancelled(self) -> None:
        """
        Run EventyApp until it is cancelled.

        This method will block until the app is cancelled.

        :return: None if the app was cancelled.
        :raises: Exception if an exception occurred processing records.
        """
        logger.info(f"EventyApp {self._app_service.name} started.")
        while not self._cancelled:
            self._process_one_batch()
        logger.info(f"EventyApp {self._app_service.name} stopped after being cancelled.")

    def run_until_end_of_topics(self) -> None:
        """
        Run EventyApp until it there is no more records to read.

        This method will block until then.

        :return: None if OK.
        :raises: Exception if an exception occurred processing records.
        """
        logger.info(f"EventyApp {self._app_service.name} started.")
        while not self._record_store.has_reached_end_of_topics():
            self._process_one_batch()
        logger.info(f"EventyApp {self._app_service.name} stopped because there is no more record to read.")

    def run(
        self,
        done_callback: Callable[[Optional[BaseException]], None] = None,
    ) -> None:
        """
        Run EventyApp in a separate thread.

        The app will be cancelled when the thread calling this method is stopped.

        If an exception is raised processing records, the callback will be called with the exception.
        If the app is otherwise cancelled, the callback will be called with None.

        :param done_callback: Callback to be called when the app is done.
        """
        calling_thread = threading.current_thread()

        def _run_then_callback():
            logger.info(f"Starting EventyApp {self._app_service.name}.")

            try:
                while not self._cancelled:
                    self._process_one_batch()
                    if not calling_thread.is_alive():
                        logger.info(f"Calling thread died. Cancelling EventyApp {self._app_service.name}.")
                        self.cancel()

            except BaseException as e:
                logger.exception(f"Error processing records in EventyApp {self._app_service.name}.")
                if done_callback:
                    done_callback(e)

            else:
                logger.info(f"EventyApp {self._app_service.name} stopped.")
                if done_callback:
                    done_callback(None)

        thread = Thread(target=_run_then_callback)
        thread.start()

    def cancel(self) -> None:
        """
        Cancel EventyApp, current batch will be processed.
        """
        logger.info(f"Cancelling EventyApp {self._app_service.name}.")
        if self._cancelled:
            logger.info(f"App was already cancelled.")
        else:
            self._cancelled = True

    def _process_one_record_one_handler(self, record: Record, handler: Handler) -> None:
        """
        Process a single record and write possible responses.

        :raises Exception: If the handler fails to process the record.
        """
        logger.debug(
            f"Processing record {record.qualified_name} data={record.data} with handler {handler}."
        )
        for output_record in handler.handle_record(record):
            topic = self._get_topic_for_record(output_record)
            logger.debug(
                f"Writing output record {output_record.qualified_name} on topic {topic} with data={output_record.data}."
            )
            self._record_store.write(output_record, topic, self._write_timeout_ms, )

    def _process_one_batch(self) -> None:
        """
        Read and process the next batch of records.

        :raises Exception: If some handler failed.
        """
        records = list(
            self._record_store.read(
                max_count=self._read_batch_size,
                timeout_ms=self._read_timeout_ms,
                auto_ack=False,
            )
        )
        if records:
            logger.debug(f"Received {len(records)} new records.")
        else:
            time.sleep(self._read_wait_time_ms / 1000)
            return

        # At Least Once
        for record in records:
            with local_trace(correlation_id=record.correlation_id):
                for handler in self._get_handlers_for_record(record, Guarantee.AT_LEAST_ONCE):
                    self._process_one_record_one_handler(record, handler)

        # Exactly Once
        self._record_store.start_transaction()
        self._record_store.ack(self._ack_timeout_ms)
        for record in records:
            with local_trace(correlation_id=record.correlation_id):
                for handler in self._get_handlers_for_record(record, Guarantee.EXACTLY_ONCE):
                    self._process_one_record_one_handler(record, handler)
        self._record_store.commit(self._commit_timeout_ms)

        # At Most Once
        for record in records:
            with local_trace(correlation_id=record.correlation_id):
                for handler in self._get_handlers_for_record(record, Guarantee.AT_MOST_ONCE):
                    self._process_one_record_one_handler(record, handler)

    def _get_service_for_name(self, service_name: Optional[str]):
        if service_name is None:
            return self._app_service
        for service in self._ext_services:
            if service.name == service_name:
                return service
        raise ValueError(f"Service {service_name} not found.")

    def _get_topics_to_register(self) -> list[str]:
        topics: set[str] = set()
        for handler in self._all_handlers:
            service = self._get_service_for_name(handler.service_name)
            topic = service.topic_for(handler.record_type)
            logger.info(f"Found handler {handler}. Will listen topic {topic}.")
            topics.add(topic)
        return list(sorted(topics))

    def _get_handlers_for_record(self, record: Record, guarantee: Guarantee) -> list[Handler]:
        handlers: list[Handler] = list()
        for handler in self._all_handlers:
            if (
                handler.record_type == record.type
                and handler.record_name == record.name
                and handler.delivery_guarantee == guarantee
                and self._get_service_for_name(handler.service_name).namespace == record.namespace
            ):
                if isinstance(record, Response):
                    if record.destination == self._app_service.namespace:
                        handlers.append(handler)
                else:
                    handlers.append(handler)
        return handlers

    def _get_topic_for_record(self, record: Record) -> str:
        if record.type == RecordType.EVENT or record.type == RecordType.RESPONSE:
            return self._app_service.topic_for(record.type)
        for service in self._ext_services:
            if service.namespace == record.namespace:
                return service.topic_for(record.type)
        raise ValueError(f"No topic for record {record}.")


class _AppRecordWriteStore(RecordWriteStore):
    """
    A RecordWriteStore implementation which delegates routing to correct topic to the EventyApp.
    """

    def __init__(self, app: EventyApp):
        self._app = app

    def write(self, record: Record) -> None:
        with local_trace(correlation_id=record.correlation_id):
            topic = self._app._get_topic_for_record(record)
            logger.debug(f"Writing record {record.qualified_name} data={record.data} to topic: {topic}.")
            self._app._record_store.write_now(
                record,
                topic,
                self._app._write_timeout_ms,
            )



# Copyright (c) Qotto, 2021
import json
import logging
import os
from datetime import datetime, timezone
from io import BytesIO
from typing import Union, Dict, Any, Optional, Iterator

import avro.schema
import semver
import yaml
from avro.datafile import DataFileWriter, DataFileReader
from avro.io import DatumWriter, AvroTypeException, DatumReader
from avro.schema import RecordSchema
from semver import VersionInfo

from eventy.record import Record, Event, Request, Response
from eventy.record.record import RecordType
from eventy.serialization import RecordSerializer
from eventy.serialization.errors import SerializationError, UnknownRecordTypeError

__all__ = [
    'AvroSerializer',
    'AvroDecodeError',
]

logger = logging.getLogger(__name__)


class AvroDecodeError(SerializationError):
    def __init__(self, decoded_data: dict) -> None:
        super().__init__(f"Failed to decode Record. Data decoded so far is: {decoded_data}")
        self.decoded_data = decoded_data


class AvroSerializer(RecordSerializer):
    """
    Avro serializer, encode and decode records in avro format.
    """

    def __init__(
        self,
    ):
        """
        Instantiate a serializer.
        """
        self.avro_schemas = {}

    def load_schemas_folder(
        self,
        schemas_folder: str, schemas_ext='.evsc.yaml', recursive=True
    ) -> None:
        """
        Parse all EVenty SChemas and register them to this AvroSerializer.
        """
        for schema_data in _load_schema_files(schemas_folder, schemas_ext, recursive):
            self.register_schema(schema_data)

    def register_schema(
        self,
        schema_data: Dict
    ) -> None:
        """
        Register an EVenty SChema so this AvroSerializer can encode eventy Records.

        :param schema_data:
        :return:
        """
        try:
            protocol_version = VersionInfo.parse(schema_data.pop('protocol_version'))
        except Exception:
            raise ValueError(f'Malformed schema_data, cannot parse protocol_version: {schema_data}.')

        try:
            namespace = schema_data.get('namespace')
            name = schema_data.get('name')
            qualified_name = f'{namespace}:{name}'
        except Exception:
            raise ValueError(f'Field namespace or name is missing: {schema_data}.')
        if qualified_name in self.avro_schemas:
            raise ValueError(f'Schema {qualified_name} already registered.')

        if protocol_version.match('>=3.0.0') and protocol_version.match('<4.0.0'):
            avro_schema = _gen_avro_schema_v3(**schema_data)
        else:
            raise ValueError(f'Cannot parse {qualified_name}, cannot handle protocol_version {protocol_version}.')

        logger.debug(f"Added avro schema for {qualified_name}.")
        self.avro_schemas[qualified_name] = avro_schema

    def encode(self, record: Record) -> bytes:
        """
        Encode an eventy Record as an avro record, with avro schema.

        :param record: An eventy Record
        :return: Encoded avro record
        """
        avro_schema = self.avro_schemas.get(record.qualified_name)
        if not avro_schema:
            raise SerializationError(f"Could not find avro schema for record {record.qualified_name}.")

        record_dict: Dict[str, Any] = {
            'record_namespace': record.namespace,
            'record_name': record.name,
            'type': record.type.value,
            'protocol_version': record.protocol_version,
            'version': record.version,
            'source': record.source,
            'uuid': record.uuid,
            'correlation_id': record.correlation_id,
            'partition_key': record.partition_key,
            'date_timestamp': int(record.date.timestamp() * 1000),
            'date_iso8601': record.date.isoformat(),
        }
        if isinstance(record, Response):
            record_dict.update(
                {
                    'destination': record.destination,
                    'request_uuid': record.request_uuid,
                    'ok': record.ok,
                    'error_code': record.error_code,
                    'error_message': record.error_message,
                }
            )
        record_dict.update(
            {
                'context': record.context,
                'data': record.data
            }
        )

        try:
            bytes_io = BytesIO()
            with DataFileWriter(bytes_io, DatumWriter(), avro_schema) as writer:
                writer.append(record_dict)
                writer.flush()
                output_bytes = bytes_io.getvalue()
            return output_bytes
        except AvroTypeException as avro_exception:
            raise SerializationError(avro_exception)

    def decode(self, encoded: bytes) -> Union[Event, Request, Response]:
        """
        Decode avro record as an eventy Record.

        Convert record encoded with protocol_version <= 3.*.*.

        :param encoded: Encoded bytes
        :return: An eventy Record
        """
        try:
            reader = DataFileReader(BytesIO(encoded), DatumReader())
            avro_schema_data = json.loads(reader.meta.get('avro.schema').decode('utf-8'))
            avro_record_data = next(reader)
        except Exception as e:
            raise SerializationError from e

        try:
            protocol_version = VersionInfo.parse(avro_record_data.get('protocol_version'))
        except Exception:
            logger.info(f"There was no protocol_version in record {avro_schema_data.get('name')}, assuming 1.0.0.")
            protocol_version = semver.VersionInfo(1)

        if protocol_version.match('<3.0.0'):
            return _decode_avro_record_legacy(avro_schema_data, avro_record_data)
        elif protocol_version.match('<4.0.0'):
            return _decode_avro_record_v3(avro_schema_data, avro_record_data)
        else:
            raise ValueError(f"Cannot handle protocol_version {protocol_version}.")

    @classmethod
    def from_schemas_folder(
        cls,
        schemas_folder: str, schemas_ext='.evsc.yaml', recursive=True
    ) -> 'AvroSerializer':
        serializer = AvroSerializer()
        serializer.load_schemas_folder(schemas_folder, schemas_ext, recursive)
        return serializer


def _decode_avro_record_legacy(
    avro_schema_data: Dict[str, str],
    avro_record_data: Dict
) -> Union[Event, Request, Response]:
    """
    Convert any avro record to a 3.0.0 compatible eventy Record.

    :param avro_schema_data: Schema data associated with avro record
    :param avro_record_data: Record data in avro record
    :return: Eventy Record
    """
    # record data for eventy Record constructor
    record_data = dict()

    try:
        # get record type
        record_type = RecordType(avro_record_data.pop('type', 'EVENT'))

        # get mandatory data
        record_data['protocol_version'] = avro_record_data.pop('protocol_version', '1.0.0')

        name = avro_schema_data.get('name') or 'Unspecified'
        record_data['name'] = name

        namespace = avro_schema_data.get('namespace') or 'urn:unspecified'
        namespace = namespace.replace('.', ':')
        if not namespace.startswith('urn:'):
            namespace = 'urn:' + namespace
        record_data['namespace'] = namespace

        record_data['version'] = avro_record_data.pop('version', '1.0.0')
        source = avro_record_data.pop('source', 'urn:unspecified').replace('.', ':')
        if not source.startswith('urn:'):
            source = 'urn:' + source
        record_data['source'] = source

        # get optional data
        for meta_key in [
            'uuid', 'correlation_id', 'partition_key', 'context',
        ]:
            if meta_key in avro_record_data:
                record_data[meta_key] = avro_record_data.pop(meta_key)

        # get date from timestamp
        for timestamp_key in [
            'timestamp', 'timestamp_ms',
            'date_timestamp', 'date_timestamp_ms',
            'event_timestamp', 'event_timestamp_ms',
        ]:
            if timestamp_key in avro_record_data:
                timestamp = avro_record_data.pop(timestamp_key)
                if 'date' not in record_data:
                    record_data['date'] = datetime.fromtimestamp(timestamp / 1000, timezone.utc)

        # get date from iso string
        for date_iso_key in [
            'date_iso', 'date_iso8601',
        ]:
            if date_iso_key in avro_record_data:
                date_iso = avro_record_data.pop(date_iso_key)
                if 'date' not in record_data:
                    record_data['date'] = datetime.fromisoformat(date_iso)

        # add data fields
        record_data['data'] = dict()
        record_data['data'].update(avro_record_data.pop('data', {}))
        record_data['data'].update(avro_record_data)

        if record_type == RecordType.EVENT:
            return Event(**record_data)
        elif record_type == RecordType.REQUEST:
            return Request(**record_data)
        elif record_type == RecordType.RESPONSE:
            return Response(**record_data)
        else:
            raise UnknownRecordTypeError(f"Unknown type: {record_type}.")
    except Exception:
        raise AvroDecodeError(decoded_data=record_data)


def _decode_avro_record_v3(
    avro_schema_data: Dict[str, str],
    avro_record_data: Dict
) -> Union[Event, Request, Response]:
    """
    Convert a 3.*.* avro record to an eventy Record.

    :param avro_schema_data: Schema data associated with avro record
    :param avro_record_data: Record data in avro record
    :return: Eventy Record
    """
    record_type = RecordType(avro_record_data.pop('type'))

    date = datetime.fromisoformat(avro_record_data.pop('date_iso8601'))
    avro_record_data.pop('date_timestamp')

    avro_record_data['date'] = date

    name = avro_record_data.pop('record_name', avro_schema_data.get('name'))
    avro_record_data['name'] = name

    namespace = avro_record_data.pop(
        'record_namespace', f'urn:{avro_schema_data.get("namespace", "unspecified").replace(".", ":")}'
    )
    avro_record_data['namespace'] = namespace

    if record_type == RecordType.EVENT:
        return Event(**avro_record_data)
    elif record_type == RecordType.REQUEST:
        return Request(**avro_record_data)
    elif record_type == RecordType.RESPONSE:
        return Response(**avro_record_data)
    else:
        raise UnknownRecordTypeError(f"Unknown type: {record_type}.")


def _gen_avro_schema_v3(
    namespace: str,
    name: str,
    type: str,
    doc: Optional[str] = None,
    data_fields: Optional[list] = None,
) -> RecordSchema:
    """
    Generate a full avro RecordSchema from an EVenty SChema.

    :param namespace: Record namespace (eventy URN format)
    :param name: Record name
    :param type: Record type
    :param doc: Record documentation
    :param data_fields: Schema of Record.data fields
    :return: An avro RecordSchema
    """
    record_type = RecordType(type.upper())

    # avro namespace format is dot-separated without urn: prefix and without -, only _ allowed
    avro_namespace = namespace.replace('urn:', '').replace(':', '.').replace('-', '_')
    avro_name = name

    fields = [
        {
            'name': 'record_namespace',
            'type': 'string',
            'doc': 'Record namespace',
        },
        {
            'name': 'record_name',
            'type': 'string',
            'doc': 'Record name',
        },
        {
            'name': 'type',
            'type': {
                'name': 'RecordType',
                'type': 'enum',
                'symbols': [record_type.value]
            },
            'doc': 'Type of of record: one of EVENT, REQUEST or RESPONSE',
        },
        {
            'name': 'protocol_version',
            'type': 'string',
            'doc': 'Version of the Eventy protocol used for encoding this message (semver format: X.Y.Z)',
        },
        {
            'name': 'version',
            'type': 'string',
            'doc': 'Version of the schema used for encoding this message (semver format: X.Y.Z)',
        },
        {
            'name': 'source', 'type': 'string',
            'doc': 'Source of the record (URN of the producing service)',
        },
        {
            'name': 'uuid', 'type': 'string', 'logicalType': 'uuid',
            'doc': 'UUID for this record',
        },
        {
            'name': 'correlation_id', 'type': 'string',
            'doc': 'Identifier propagated across the system and to link associated records together',
        },
        {
            'name': 'partition_key', 'type': ['string', 'null'],
            'doc': 'A string determining to which partition your record will be assigned',
        },
        {
            'name': 'date_timestamp', 'type': 'long', 'logicalType': 'timestamp-millis',
            'doc': 'UNIX timestamp in milliseconds',
        },
        {
            'name': 'date_iso8601', 'type': 'string',
            'doc': 'ISO 8601 date with timezone',
        },
    ]
    if record_type == RecordType.RESPONSE:
        fields += [
            {
                'name': 'destination', 'type': 'string',
                'doc': 'URN of the destination service',
            },
            {
                'name': 'request_uuid', 'type': 'string',
                'doc': 'UUID of the associated request',
            },
            {
                'name': 'ok', 'type': 'boolean',
                'doc': 'Status: True if there was no error, false otherwise',
            },
            {
                'name': 'error_code', 'type': ['null', 'int'],
                'doc': 'Numeric code for the error, if ok is False, null otherwise',
            },
            {
                'name': 'error_message', 'type': ['null', 'string'],
                'doc': 'Description for the error, if ok is False, null otherwise',
            },
        ]
    context_field = {
        'name': 'context', 'type':
            {
                'type': 'map', 'values':
                {
                    'type': 'map',
                    'values': ['string', 'double', 'float', 'long', 'int', 'boolean', 'null']
                }
            },
        'doc': 'Context data, always propagated'
    }
    data_field: Dict[str, Any] = {
        'name': 'data', 'type':
            {
                'type': 'record', 'name': f'{avro_name}_data', 'namespace': avro_namespace, 'fields':
                data_fields or []
            },
        'doc': 'Record payload',
    }
    if record_type == RecordType.RESPONSE:
        data_field = {
            'name': 'data', 'type': [
                {
                    'type': 'map',
                    'values': []
                },
                {
                    'type': 'record', 'name': f'{avro_name}_data', 'namespace': avro_namespace, 'fields':
                    data_fields or []
                },
            ],
            'doc': 'Record payload',
        }

    fields += [context_field, data_field, ]
    schema_dict = {
        'name': avro_name,
        'namespace': avro_namespace,
        'doc': doc or f'{name} {record_type} Record',
        'type': 'record',
        'fields': fields,
    }
    return avro.schema.parse(json.dumps(schema_dict))


def _load_schema_files(
    schemas_folder: str,
    schemas_ext: str,
    recursive: bool,
) -> Iterator[Dict]:
    with os.scandir(schemas_folder) as entries:
        entry: os.DirEntry
        for entry in entries:

            # load in sub folder if recursive option set
            if entry.is_dir():
                if recursive:
                    yield from _load_schema_files(entry.path, schemas_ext, recursive)
                else:
                    continue

            # ignores everything that is not a file
            if not entry.is_file():
                continue  # pragma: nocover

            # ignores hidden files
            if entry.name.startswith("."):
                continue  # pragma: nocover

            # ignores everything files with wrong extension
            if not entry.name.endswith(schemas_ext):
                continue  # pragma: nocover

            with open(entry.path) as avro_yml:
                yaml_data = yaml.load(avro_yml.read(), Loader=yaml.SafeLoader)
                yield yaml_data
